use once_cell::sync::Lazy;
use std::collections::BTreeMap;
use std::collections::BTreeSet;
use std::iter::FromIterator;

pub(crate) static BUILTIN_IDENTS: Lazy<BTreeSet<&'static str>> = Lazy::new(|| {
    BTreeSet::from_iter(vec![
        "str",
        "char",
        "bool",
        "JsValue",
        "u8",
        "i8",
        "u16",
        "i16",
        "u32",
        "i32",
        "u64",
        "i64",
        "usize",
        "isize",
        "f32",
        "f64",
        "Result",
        "String",
        "Vec",
        "Option",
        "Array",
        "ArrayBuffer",
        "Object",
        "Promise",
        "Function",
        "Clamped",
        "DataView",
        "Iterator",
        "AsyncIterator",
    ])
});

// whitelist a few names that have known polyfills
pub(crate) static POLYFILL_INTERFACES: Lazy<BTreeSet<&'static str>> =
    Lazy::new(|| BTreeSet::from_iter(vec!["AudioContext", "OfflineAudioContext"]));

pub(crate) static IMMUTABLE_SLICE_WHITELIST: Lazy<BTreeSet<&'static str>> = Lazy::new(|| {
    BTreeSet::from_iter(vec![
        // ImageData
        "ImageData",
        // WebGlRenderingContext, WebGl2RenderingContext
        "uniform1fv",
        "uniform2fv",
        "uniform3fv",
        "uniform4fv",
        "uniform1iv",
        "uniform2iv",
        "uniform3iv",
        "uniform4iv",
        "uniformMatrix2fv",
        "uniformMatrix3fv",
        "uniformMatrix4fv",
        "uniformMatrix2x3fv",
        "uniformMatrix2x4fv",
        "uniformMatrix3x2fv",
        "uniformMatrix3x4fv",
        "uniformMatrix4x2fv",
        "uniformMatrix4x3fv",
        "vertexAttrib1fv",
        "vertexAttrib2fv",
        "vertexAttrib3fv",
        "vertexAttrib4fv",
        "bufferData",
        "bufferSubData",
        "texImage2D",
        "texSubImage2D",
        "compressedTexImage2D",
        // WebGl2RenderingContext
        "uniform1uiv",
        "uniform2uiv",
        "uniform3uiv",
        "uniform4uiv",
        "texImage3D",
        "texSubImage3D",
        "compressedTexImage3D",
        "clearBufferfv",
        "clearBufferiv",
        "clearBufferuiv",
        // WebSocket
        "send",
        // WebGPU
        "setBindGroup",
        "writeBuffer",
        "writeTexture",
        // AudioBuffer
        "copyToChannel",
        // FontFace
        "FontFace", // TODO: Add another type's functions here. Leave a comment header with the type name
        // FileSystemSyncAccessHandle and FileSystemWritableFileStream
        "write",
        // SubtleCrypto
        "encrypt",
        "decrypt",
        "digest",
        "sign",
        "unwrapKey",
        "verify",
    ])
});

pub(crate) static FIXED_INTERFACES: Lazy<
    BTreeMap<&'static str, BTreeMap<&'static str, &'static str>>,
> = Lazy::new(|| {
    let image_bitmap = BTreeMap::from_iter([
        ("create_image_bitmap_with_html_image_element_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_html_image_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_svg_image_element_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_svg_image_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_html_canvas_element_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_html_canvas_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_html_video_element_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_html_video_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_image_bitmap_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_image_bitmap_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_offscreen_canvas_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_offscreen_canvas_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_video_frame_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_video_frame_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_blob_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_blob_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_image_data_and_i32_and_a_sy_and_a_sw_and_a_sh", "create_image_bitmap_with_image_data_and_a_sx_and_a_sy_and_a_sw_and_a_sh"),
        ("create_image_bitmap_with_html_image_element_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_html_image_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_svg_image_element_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_svg_image_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_html_canvas_element_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_html_canvas_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_html_video_element_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_html_video_element_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_image_bitmap_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_image_bitmap_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_offscreen_canvas_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_offscreen_canvas_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_video_frame_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_video_frame_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_blob_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_blob_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
        ("create_image_bitmap_with_image_data_and_i32_and_a_sy_and_a_sw_and_a_sh_and_a_options", "create_image_bitmap_with_image_data_and_a_sx_and_a_sy_and_a_sw_and_a_sh_and_a_options"),
    ]);

    let canvas_rendering_context = BTreeMap::from_iter([
        (
            "set_transform",
            "set_transform_with_default_dom_matrix_2d_init",
        ),
        (
            "set_transform_with_a_and_b_and_c_and_d_and_e_and_f",
            "set_transform",
        ),
    ]);

    BTreeMap::from_iter([
        ("Window", image_bitmap.clone()),
        ("WorkerGlobalScope", image_bitmap),
        ("CanvasRenderingContext2d", canvas_rendering_context.clone()),
        (
            "OffscreenCanvasRenderingContext2d",
            canvas_rendering_context,
        ),
        (
            "ReadableStreamByobReader",
            BTreeMap::from_iter([("read", "read_with_array_buffer_view")]),
        ),
        (
            "AudioBufferSourceNode",
            BTreeMap::from_iter([
                (
                    "start_with_when_and_offset",
                    "start_with_when_and_grain_offset",
                ),
                (
                    "start_with_when_and_offset_and_duration",
                    "start_with_when_and_grain_offset_and_grain_duration",
                ),
            ]),
        ),
        (
            "IdbFactory",
            BTreeMap::from_iter([("open_with_options", "open_with_idb_open_db_options")]),
        ),
        (
            "IdbDatabase",
            BTreeMap::from_iter([(
                "create_object_store_with_options",
                "create_object_store_with_optional_parameters",
            )]),
        ),
        (
            "IdbObjectStore",
            BTreeMap::from_iter([
                ("count_with_query", "count_with_key"),
                (
                    "create_index_with_str_and_options",
                    "create_index_with_str_and_optional_parameters",
                ),
                (
                    "create_index_with_str_sequence_and_options",
                    "create_index_with_str_sequence_and_optional_parameters",
                ),
                ("open_cursor_with_query", "open_cursor_with_range"),
                (
                    "open_cursor_with_query_and_direction",
                    "open_cursor_with_range_and_direction",
                ),
                ("open_key_cursor_with_query", "open_key_cursor_with_range"),
                (
                    "open_key_cursor_with_query_and_direction",
                    "open_key_cursor_with_range_and_direction",
                ),
                ("get_all_with_query", "get_all_with_key"),
                ("get_all_with_query_and_count", "get_all_with_key_and_limit"),
                ("get_all_keys_with_query", "get_all_keys_with_key"),
                (
                    "get_all_keys_with_query_and_count",
                    "get_all_keys_with_key_and_limit",
                ),
            ]),
        ),
        (
            "IdbIndex",
            BTreeMap::from_iter([
                ("count_with_query", "count_with_key"),
                ("open_cursor_with_query", "open_cursor_with_range"),
                (
                    "open_cursor_with_query_and_direction",
                    "open_cursor_with_range_and_direction",
                ),
                ("open_key_cursor_with_query", "open_key_cursor_with_range"),
                (
                    "open_key_cursor_with_query_and_direction",
                    "open_key_cursor_with_range_and_direction",
                ),
                ("get_all_with_query", "get_all_with_key"),
                ("get_all_with_query_and_count", "get_all_with_key_and_limit"),
                ("get_all_keys_with_query", "get_all_keys_with_key"),
                (
                    "get_all_keys_with_query_and_count",
                    "get_all_keys_with_key_and_limit",
                ),
            ]),
        ),
        (
            "Navigator",
            BTreeMap::from_iter([
                ("vibrate_with_u32", "vibrate_with_duration"),
                ("vibrate_with_u32_sequence", "vibrate_with_pattern"),
            ]),
        ),
    ])
});

pub(crate) static BREAKING_GETTER_THROWS: Lazy<BTreeMap<&'static str, Vec<&'static str>>> =
    Lazy::new(|| {
        BTreeMap::from_iter([
            ("Blob", vec!["size"]),
            ("CanvasRenderingContext2D", vec!["lineJoin"]),
            ("ExtendableMessageEvent", vec!["data"]),
            ("File", vec!["lastModified"]),
            ("FileSystemEntry", vec!["name", "fullPath"]),
            ("HTMLElement", vec!["innerText"]),
            ("MessageEvent", vec!["data"]),
            ("OffscreenCanvasRenderingContext2D", vec!["lineJoin"]),
            ("URL", vec!["origin"]),
        ])
    });

pub(crate) static BREAKING_SETTER_THROWS: Lazy<BTreeMap<&'static str, Vec<&'static str>>> =
    Lazy::new(|| {
        BTreeMap::from_iter([
            (
                "AnalyserNode",
                vec![
                    "fftSize",
                    "minDecibels",
                    "maxDecibels",
                    "smoothingTimeConstant",
                ],
            ),
            ("Animation", vec!["currentTime"]),
            ("Attr", vec!["value"]),
            (
                "AudioNode",
                vec!["channelCount", "channelCountMode", "channelInterpretation"],
            ),
            ("CSSConditionRule", vec!["conditionText"]),
            ("CSSFontFeatureValuesRule", vec!["fontFamily", "valueText"]),
            ("CSSStyleDeclaration", vec!["cssText"]),
            ("CanvasRenderingContext2D", vec!["filter", "font"]),
            ("CharacterData", vec!["data"]),
            ("ConvolverNode", vec!["buffer"]),
            ("DOMTokenList", vec!["value"]),
            ("Document", vec!["title", "body"]),
            ("Element", vec!["innerHTML", "outerHTML", "slot"]),
            (
                "FontFace",
                vec![
                    "family",
                    "style",
                    "weight",
                    "stretch",
                    "unicodeRange",
                    "variant",
                    "featureSettings",
                    "variationSettings",
                    "display",
                ],
            ),
            (
                "HTMLAnchorElement",
                vec![
                    "target",
                    "download",
                    "ping",
                    "rel",
                    "referrerPolicy",
                    "hreflang",
                    "type",
                    "coords",
                    "charset",
                    "name",
                    "rev",
                    "shape",
                    "href",
                ],
            ),
            (
                "HTMLAreaElement",
                vec![
                    "alt",
                    "coords",
                    "shape",
                    "target",
                    "download",
                    "ping",
                    "rel",
                    "referrerPolicy",
                    "noHref",
                    "href",
                ],
            ),
            ("HTMLBRElement", vec!["clear"]),
            ("HTMLBaseElement", vec!["href", "target"]),
            (
                "HTMLBodyElement",
                vec!["text", "link", "vLink", "aLink", "bgColor", "background"],
            ),
            (
                "HTMLButtonElement",
                vec![
                    "autofocus",
                    "disabled",
                    "formAction",
                    "formEnctype",
                    "formMethod",
                    "formNoValidate",
                    "formTarget",
                    "name",
                    "type",
                    "value",
                ],
            ),
            ("HTMLCanvasElement", vec!["width", "height"]),
            ("HTMLDListElement", vec!["compact"]),
            ("HTMLDataElement", vec!["value"]),
            ("HTMLDetailsElement", vec!["open"]),
            ("HTMLDialogElement", vec!["open"]),
            ("HTMLDirectoryElement", vec!["compact"]),
            ("HTMLDivElement", vec!["align"]),
            ("HTMLDocument", vec!["domain", "designMode"]),
            (
                "HTMLElement",
                vec![
                    "dir",
                    "hidden",
                    "tabIndex",
                    "accessKey",
                    "draggable",
                    "contentEditable",
                    "spellcheck",
                ],
            ),
            (
                "HTMLEmbedElement",
                vec!["src", "type", "width", "height", "align", "name"],
            ),
            ("HTMLFieldSetElement", vec!["disabled", "name"]),
            ("HTMLFontElement", vec!["color", "face", "size"]),
            (
                "HTMLFormElement",
                vec![
                    "acceptCharset",
                    "action",
                    "autocomplete",
                    "enctype",
                    "encoding",
                    "method",
                    "name",
                    "noValidate",
                    "target",
                ],
            ),
            (
                "HTMLFrameElement",
                vec![
                    "name",
                    "scrolling",
                    "src",
                    "frameBorder",
                    "longDesc",
                    "noResize",
                    "marginHeight",
                    "marginWidth",
                ],
            ),
            ("HTMLFrameSetElement", vec!["cols", "rows"]),
            (
                "HTMLHRElement",
                vec!["align", "color", "noShade", "size", "width"],
            ),
            ("HTMLHeadingElement", vec!["align"]),
            ("HTMLHtmlElement", vec!["version"]),
            (
                "HTMLIFrameElement",
                vec![
                    "src",
                    "srcdoc",
                    "name",
                    "allowFullscreen",
                    "allowPaymentRequest",
                    "width",
                    "height",
                    "referrerPolicy",
                    "align",
                    "scrolling",
                    "frameBorder",
                    "longDesc",
                    "marginHeight",
                    "marginWidth",
                ],
            ),
            (
                "HTMLImageElement",
                vec![
                    "alt",
                    "src",
                    "srcset",
                    "crossOrigin",
                    "useMap",
                    "referrerPolicy",
                    "isMap",
                    "width",
                    "height",
                    "decoding",
                    "name",
                    "align",
                    "hspace",
                    "vspace",
                    "longDesc",
                    "border",
                    "sizes",
                ],
            ),
            (
                "HTMLInputElement",
                vec![
                    "accept",
                    "alt",
                    "autocomplete",
                    "autofocus",
                    "defaultChecked",
                    "disabled",
                    "formAction",
                    "formEnctype",
                    "formMethod",
                    "formNoValidate",
                    "formTarget",
                    "height",
                    "inputMode",
                    "max",
                    "maxLength",
                    "min",
                    "minLength",
                    "multiple",
                    "name",
                    "pattern",
                    "placeholder",
                    "readOnly",
                    "required",
                    "size",
                    "src",
                    "step",
                    "type",
                    "defaultValue",
                    "value",
                    "valueAsNumber",
                    "width",
                    "align",
                    "useMap",
                    "webkitdirectory",
                ],
            ),
            ("HTMLLIElement", vec!["value", "type"]),
            ("HTMLLegendElement", vec!["align"]),
            (
                "HTMLLinkElement",
                vec![
                    "href",
                    "crossOrigin",
                    "rel",
                    "media",
                    "hreflang",
                    "type",
                    "referrerPolicy",
                    "charset",
                    "rev",
                    "target",
                    "integrity",
                    "as",
                ],
            ),
            ("HTMLMapElement", vec!["name"]),
            (
                "HTMLMediaElement",
                vec![
                    "src",
                    "crossOrigin",
                    "preload",
                    "currentTime",
                    "defaultPlaybackRate",
                    "playbackRate",
                    "autoplay",
                    "loop",
                    "controls",
                    "volume",
                    "defaultMuted",
                ],
            ),
            ("HTMLMenuElement", vec!["type", "label", "compact"]),
            (
                "HTMLMenuItemElement",
                vec![
                    "type",
                    "label",
                    "icon",
                    "disabled",
                    "radiogroup",
                    "defaultChecked",
                ],
            ),
            (
                "HTMLMetaElement",
                vec!["name", "httpEquiv", "content", "scheme"],
            ),
            (
                "HTMLMeterElement",
                vec!["value", "min", "max", "low", "high", "optimum"],
            ),
            ("HTMLModElement", vec!["cite", "dateTime"]),
            (
                "HTMLOListElement",
                vec!["reversed", "start", "type", "compact"],
            ),
            (
                "HTMLObjectElement",
                vec![
                    "data",
                    "type",
                    "typeMustMatch",
                    "name",
                    "useMap",
                    "width",
                    "height",
                    "align",
                    "archive",
                    "code",
                    "declare",
                    "hspace",
                    "standby",
                    "vspace",
                    "codeBase",
                    "codeType",
                    "border",
                ],
            ),
            ("HTMLOptGroupElement", vec!["disabled", "label"]),
            (
                "HTMLOptionElement",
                vec!["disabled", "label", "defaultSelected", "value", "text"],
            ),
            ("HTMLOptionsCollection", vec!["length"]),
            ("HTMLOutputElement", vec!["name", "defaultValue", "value"]),
            ("HTMLParagraphElement", vec!["align"]),
            (
                "HTMLParamElement",
                vec!["name", "value", "type", "valueType"],
            ),
            ("HTMLPreElement", vec!["width"]),
            ("HTMLProgressElement", vec!["value", "max"]),
            ("HTMLQuoteElement", vec!["cite"]),
            (
                "HTMLScriptElement",
                vec![
                    "src",
                    "type",
                    "noModule",
                    "charset",
                    "async",
                    "defer",
                    "crossOrigin",
                    "event",
                    "htmlFor",
                    "integrity",
                ],
            ),
            (
                "HTMLSelectElement",
                vec![
                    "autofocus",
                    "autocomplete",
                    "disabled",
                    "multiple",
                    "name",
                    "required",
                    "size",
                    "length",
                    "selectedIndex",
                ],
            ),
            ("HTMLSlotElement", vec!["name"]),
            (
                "HTMLSourceElement",
                vec!["src", "type", "srcset", "sizes", "media"],
            ),
            ("HTMLStyleElement", vec!["media", "type"]),
            ("HTMLTableCaptionElement", vec!["align"]),
            (
                "HTMLTableCellElement",
                vec![
                    "colSpan", "rowSpan", "headers", "align", "axis", "height", "width", "ch",
                    "chOff", "noWrap", "vAlign", "bgColor",
                ],
            ),
            (
                "HTMLTableColElement",
                vec!["span", "align", "ch", "chOff", "vAlign", "width"],
            ),
            (
                "HTMLTableElement",
                vec![
                    "caption",
                    "tHead",
                    "tFoot",
                    "align",
                    "border",
                    "frame",
                    "rules",
                    "summary",
                    "width",
                    "bgColor",
                    "cellPadding",
                    "cellSpacing",
                ],
            ),
            (
                "HTMLTableRowElement",
                vec!["align", "ch", "chOff", "vAlign", "bgColor"],
            ),
            (
                "HTMLTableSectionElement",
                vec!["align", "ch", "chOff", "vAlign"],
            ),
            (
                "HTMLTextAreaElement",
                vec![
                    "autocomplete",
                    "autofocus",
                    "cols",
                    "disabled",
                    "maxLength",
                    "minLength",
                    "name",
                    "placeholder",
                    "readOnly",
                    "required",
                    "rows",
                    "wrap",
                    "value",
                ],
            ),
            ("HTMLTimeElement", vec!["dateTime"]),
            (
                "HTMLTrackElement",
                vec!["kind", "src", "srclang", "label", "default"],
            ),
            ("HTMLUListElement", vec!["compact", "type"]),
            ("HTMLVideoElement", vec!["width", "height", "poster"]),
            ("IDBIndex", vec!["name"]),
            ("IDBObjectStore", vec!["name"]),
            ("MediaSource", vec!["duration"]),
            ("Node", vec!["nodeValue", "textContent"]),
            ("OffscreenCanvas", vec!["width", "height"]),
            ("OffscreenCanvasRenderingContext2D", vec!["filter", "font"]),
            ("OscillatorNode", vec!["type"]),
            (
                "SVGAElement",
                vec![
                    "download",
                    "ping",
                    "rel",
                    "referrerPolicy",
                    "hreflang",
                    "type",
                ],
            ),
            (
                "SVGAngle",
                vec!["value", "valueInSpecifiedUnits", "valueAsString"],
            ),
            ("SVGAnimatedEnumeration", vec!["baseVal"]),
            ("SVGElement", vec!["tabIndex"]),
            ("SVGLength", vec!["valueInSpecifiedUnits", "valueAsString"]),
            ("SVGMatrix", vec!["a", "b", "c", "d", "e", "f"]),
            ("SVGNumber", vec!["value"]),
            (
                "SVGPathSegArcAbs",
                vec!["x", "y", "r1", "r2", "angle", "largeArcFlag", "sweepFlag"],
            ),
            (
                "SVGPathSegArcRel",
                vec!["x", "y", "r1", "r2", "angle", "largeArcFlag", "sweepFlag"],
            ),
            (
                "SVGPathSegCurvetoCubicAbs",
                vec!["x", "y", "x1", "y1", "x2", "y2"],
            ),
            (
                "SVGPathSegCurvetoCubicRel",
                vec!["x", "y", "x1", "y1", "x2", "y2"],
            ),
            (
                "SVGPathSegCurvetoCubicSmoothAbs",
                vec!["x", "y", "x2", "y2"],
            ),
            (
                "SVGPathSegCurvetoCubicSmoothRel",
                vec!["x", "y", "x2", "y2"],
            ),
            ("SVGPathSegCurvetoQuadraticAbs", vec!["x", "y", "x1", "y1"]),
            ("SVGPathSegCurvetoQuadraticRel", vec!["x", "y", "x1", "y1"]),
            ("SVGPathSegCurvetoQuadraticSmoothAbs", vec!["x", "y"]),
            ("SVGPathSegCurvetoQuadraticSmoothRel", vec!["x", "y"]),
            ("SVGPathSegLinetoAbs", vec!["x", "y"]),
            ("SVGPathSegLinetoHorizontalAbs", vec!["x"]),
            ("SVGPathSegLinetoHorizontalRel", vec!["x"]),
            ("SVGPathSegLinetoRel", vec!["x", "y"]),
            ("SVGPathSegLinetoVerticalAbs", vec!["y"]),
            ("SVGPathSegLinetoVerticalRel", vec!["y"]),
            ("SVGPathSegMovetoAbs", vec!["x", "y"]),
            ("SVGPathSegMovetoRel", vec!["x", "y"]),
            ("SVGPoint", vec!["x", "y"]),
            ("SVGPreserveAspectRatio", vec!["align", "meetOrSlice"]),
            ("SVGRect", vec!["x", "y", "width", "height"]),
            ("SVGSVGElement", vec!["zoomAndPan"]),
            ("SVGScriptElement", vec!["type", "crossOrigin"]),
            (
                "SVGStyleElement",
                vec!["xmlspace", "type", "media", "title"],
            ),
            ("SVGViewElement", vec!["zoomAndPan"]),
            ("SVGZoomAndPan", vec!["zoomAndPan"]),
            ("ShadowRoot", vec!["innerHTML"]),
            (
                "SourceBuffer",
                vec![
                    "mode",
                    "timestampOffset",
                    "appendWindowStart",
                    "appendWindowEnd",
                ],
            ),
            ("TreeWalker", vec!["currentNode"]),
            ("URL", vec!["href", "protocol"]),
            (
                "VTTCue",
                vec!["lineAlign", "position", "positionAlign", "size"],
            ),
            (
                "VTTRegion",
                vec![
                    "width",
                    "lines",
                    "regionAnchorX",
                    "regionAnchorY",
                    "viewportAnchorX",
                    "viewportAnchorY",
                ],
            ),
            ("WaveShaperNode", vec!["curve"]),
            (
                "XMLHttpRequest",
                vec!["timeout", "withCredentials", "responseType"],
            ),
        ])
    });
